@@ -15,8 +15,6 @@
 #include <ctype.h>
 #include <math.h>
 
-#undef NDEBUG
-#define NDEBUG
 #include "id.h"
 #include "internal.h"
 #include "internal/array.h"
@@ -36,10 +34,6 @@
 #else
 static VALUE RFLOAT_0;
 #endif
-#if defined(HAVE_SIGNBIT) && defined(__GNUC__) && defined(__sun) && \
-    !defined(signbit)
-extern int signbit(double);
-#endif
 
 VALUE rb_cComplex;
 
@@ -56,8 +50,6 @@ static ID id_abs, id_arg,
 #define id_quo idQuo
 #define id_fdiv idFdiv
 
-#define f_boolcast(x) ((x) ? Qtrue : Qfalse)
-
 #define fun1(n) \
 inline static VALUE \
 f_##n(VALUE x)\
@@ -342,36 +334,36 @@ f_zero_p(VALUE x)
 
 #define f_nonzero_p(x) (!f_zero_p(x))
 
-VALUE rb_flo_is_finite_p(VALUE num);
+static inline bool
+always_finite_type_p(VALUE x)
+{
+    if (FIXNUM_P(x)) return true;
+    if (FLONUM_P(x)) return true; /* Infinity can't be a flonum */
+    return (RB_INTEGER_TYPE_P(x) || RB_TYPE_P(x, T_RATIONAL));
+}
+
 inline static int
 f_finite_p(VALUE x)
 {
-    if (RB_INTEGER_TYPE_P(x)) {
+    if (always_finite_type_p(x)) {
         return TRUE;
     }
     else if (RB_FLOAT_TYPE_P(x)) {
-	return (int)rb_flo_is_finite_p(x);
-    }
-    else if (RB_TYPE_P(x, T_RATIONAL)) {
-	return TRUE;
+	return isfinite(RFLOAT_VALUE(x));
     }
     return RTEST(rb_funcallv(x, id_finite_p, 0, 0));
 }
 
-VALUE rb_flo_is_infinite_p(VALUE num);
-inline static VALUE
+inline static int
 f_infinite_p(VALUE x)
 {
-    if (RB_INTEGER_TYPE_P(x)) {
-        return Qnil;
+    if (always_finite_type_p(x)) {
+        return FALSE;
     }
     else if (RB_FLOAT_TYPE_P(x)) {
-	return rb_flo_is_infinite_p(x);
+	return isinf(RFLOAT_VALUE(x));
     }
-    else if (RB_TYPE_P(x, T_RATIONAL)) {
-        return Qnil;
-    }
-    return rb_funcallv(x, id_infinite_p, 0, 0);
+    return RTEST(rb_funcallv(x, id_infinite_p, 0, 0));
 }
 
 inline static int
@@ -429,18 +421,6 @@ f_complex_new_bang2(VALUE klass, VALUE x, VALUE y)
     return nucomp_s_new_internal(klass, x, y);
 }
 
-#ifdef CANONICALIZATION_FOR_MATHN
-static int canonicalization = 0;
-
-RUBY_FUNC_EXPORTED void
-nucomp_canonicalization(int f)
-{
-    canonicalization = f;
-}
-#else
-#define canonicalization 0
-#endif
-
 inline static void
 nucomp_real_check(VALUE num)
 {
@@ -456,10 +436,6 @@ inline static VALUE
 nucomp_s_canonicalize_internal(VALUE klass, VALUE real, VALUE imag)
 {
     int complex_r, complex_i;
-#ifdef CANONICALIZATION_FOR_MATHN
-    if (k_exact_zero_p(imag) && canonicalization)
-	return real;
-#endif
     complex_r = RB_TYPE_P(real, T_COMPLEX);
     complex_i = RB_TYPE_P(imag, T_COMPLEX);
     if (!complex_r && !complex_i) {
@@ -636,14 +612,12 @@ f_complex_polar(VALUE klass, VALUE x, VALUE y)
     assert(!RB_TYPE_P(x, T_COMPLEX));
     assert(!RB_TYPE_P(y, T_COMPLEX));
     if (f_zero_p(x) || f_zero_p(y)) {
-	if (canonicalization) return x;
 	return nucomp_s_new_internal(klass, x, RFLOAT_0);
     }
     if (RB_FLOAT_TYPE_P(y)) {
 	const double arg = RFLOAT_VALUE(y);
 	if (arg == M_PI) {
 	    x = f_negate(x);
-	    if (canonicalization) return x;
 	    y = RFLOAT_0;
 	}
 	else if (arg == M_PI_2) {
@@ -658,14 +632,12 @@ f_complex_polar(VALUE klass, VALUE x, VALUE y)
 	    const double abs = RFLOAT_VALUE(x);
 	    const double real = abs * cos(arg), imag = abs * sin(arg);
 	    x = DBL2NUM(real);
-	    if (canonicalization && imag == 0.0) return x;
 	    y = DBL2NUM(imag);
 	}
 	else {
             const double ax = sin(arg), ay = cos(arg);
             y = f_mul(x, DBL2NUM(ax));
             x = f_mul(x, DBL2NUM(ay));
-	    if (canonicalization && f_zero_p(y)) return x;
 	}
 	return nucomp_s_new_internal(klass, x, y);
     }
@@ -725,7 +697,6 @@ nucomp_s_polar(int argc, VALUE *argv, VALUE klass)
     switch (rb_scan_args(argc, argv, "11", &abs, &arg)) {
       case 1:
 	nucomp_real_check(abs);
-	if (canonicalization) return abs;
 	return nucomp_s_new_internal(klass, abs, ZERO);
       default:
 	nucomp_real_check(abs);
@@ -1085,7 +1056,7 @@ rb_complex_pow(VALUE self, VALUE other)
     if (k_numeric_p(other) && f_real_p(other)) {
 	VALUE r, theta;
 
-	if (RB_TYPE_P(other, T_BIGNUM))
+	if (RB_BIGNUM_TYPE_P(other))
 	    rb_warn("in a**b, b may be too big");
 
 	r = f_abs(self);
@@ -1115,15 +1086,15 @@ nucomp_eqeq_p(VALUE self, VALUE other)
     if (RB_TYPE_P(other, T_COMPLEX)) {
 	get_dat2(self, other);
 
-	return f_boolcast(f_eqeq_p(adat->real, bdat->real) &&
+	return RBOOL(f_eqeq_p(adat->real, bdat->real) &&
 			  f_eqeq_p(adat->imag, bdat->imag));
     }
     if (k_numeric_p(other) && f_real_p(other)) {
 	get_dat1(self);
 
-	return f_boolcast(f_eqeq_p(dat->real, other) && f_zero_p(dat->imag));
+	return RBOOL(f_eqeq_p(dat->real, other) && f_zero_p(dat->imag));
     }
-    return f_boolcast(f_eqeq_p(other, self));
+    return RBOOL(f_eqeq_p(other, self));
 }
 
 static bool
@@ -1295,7 +1266,7 @@ rb_complex_conjugate(VALUE self)
  * Returns false, even if the complex number has no imaginary part.
  */
 static VALUE
-nucomp_false(VALUE self)
+nucomp_real_p_m(VALUE self)
 {
     return Qfalse;
 }
@@ -1349,8 +1320,8 @@ nucomp_numerator(VALUE self)
 }
 
 /* :nodoc: */
-static VALUE
-nucomp_hash(VALUE self)
+st_index_t
+rb_complex_hash(VALUE self)
 {
     st_index_t v, h[2];
     VALUE n;
@@ -1361,7 +1332,13 @@ nucomp_hash(VALUE self)
     n = rb_hash(dat->imag);
     h[1] = NUM2LONG(n);
     v = rb_memhash(h, sizeof(h));
-    return ST2FIX(v);
+    return v;
+}
+
+static VALUE
+nucomp_hash(VALUE self)
+{
+    return ST2FIX(rb_complex_hash(self));
 }
 
 /* :nodoc: */
@@ -1371,7 +1348,7 @@ nucomp_eql_p(VALUE self, VALUE other)
     if (RB_TYPE_P(other, T_COMPLEX)) {
 	get_dat2(self, other);
 
-	return f_boolcast((CLASS_OF(adat->real) == CLASS_OF(bdat->real)) &&
+	return RBOOL((CLASS_OF(adat->real) == CLASS_OF(bdat->real)) &&
 			  (CLASS_OF(adat->imag) == CLASS_OF(bdat->imag)) &&
 			  f_eqeq_p(self, other));
 
@@ -1472,10 +1449,7 @@ rb_complex_finite_p(VALUE self)
 {
     get_dat1(self);
 
-    if (f_finite_p(dat->real) && f_finite_p(dat->imag)) {
-	return Qtrue;
-    }
-    return Qfalse;
+    return RBOOL(f_finite_p(dat->real) && f_finite_p(dat->imag));
 }
 
 /*
@@ -1495,7 +1469,7 @@ rb_complex_infinite_p(VALUE self)
 {
     get_dat1(self);
 
-    if (NIL_P(f_infinite_p(dat->real)) && NIL_P(f_infinite_p(dat->imag))) {
+    if (!f_infinite_p(dat->real) && !f_infinite_p(dat->imag)) {
 	return Qnil;
     }
     return ONE;
@@ -1545,8 +1519,6 @@ nucomp_marshal_load(VALUE self, VALUE a)
     return self;
 }
 
-/* --- */
-
 VALUE
 rb_complex_raw(VALUE x, VALUE y)
 {
@@ -1580,13 +1552,6 @@ rb_Complex(VALUE x, VALUE y)
     return nucomp_s_convert(2, a, rb_cComplex);
 }
 
-/*!
- * Creates a Complex object.
- *
- * \param real    real part value
- * \param imag    imaginary part value
- * \return        a new Complex object
- */
 VALUE
 rb_dbl_complex_new(double real, double imag)
 {
@@ -2169,8 +2134,6 @@ nucomp_s_convert(int argc, VALUE *argv, VALUE klass)
     return nucomp_convert(klass, a1, a2, TRUE);
 }
 
-/* --- */
-
 /*
  * call-seq:
  *    num.real  ->  self
@@ -2237,8 +2200,6 @@ numeric_rect(VALUE self)
     return rb_assoc_new(self, INT2FIX(0));
 }
 
-static VALUE float_arg(VALUE self);
-
 /*
  * call-seq:
  *    num.polar  ->  array
@@ -2337,20 +2298,17 @@ void
 Init_Complex(void)
 {
     VALUE compat;
-#undef rb_intern
-#define rb_intern(str) rb_intern_const(str)
-
-    id_abs = rb_intern("abs");
-    id_arg = rb_intern("arg");
-    id_denominator = rb_intern("denominator");
-    id_numerator = rb_intern("numerator");
-    id_real_p = rb_intern("real?");
-    id_i_real = rb_intern("@real");
-    id_i_imag = rb_intern("@image"); /* @image, not @imag */
-    id_finite_p = rb_intern("finite?");
-    id_infinite_p = rb_intern("infinite?");
-    id_rationalize = rb_intern("rationalize");
-    id_PI = rb_intern("PI");
+    id_abs = rb_intern_const("abs");
+    id_arg = rb_intern_const("arg");
+    id_denominator = rb_intern_const("denominator");
+    id_numerator = rb_intern_const("numerator");
+    id_real_p = rb_intern_const("real?");
+    id_i_real = rb_intern_const("@real");
+    id_i_imag = rb_intern_const("@image"); /* @image, not @imag */
+    id_finite_p = rb_intern_const("finite?");
+    id_infinite_p = rb_intern_const("infinite?");
+    id_rationalize = rb_intern_const("rationalize");
+    id_PI = rb_intern_const("PI");
 
     rb_cComplex = rb_define_class("Complex", rb_cNumeric);
 
@@ -2407,7 +2365,7 @@ Init_Complex(void)
     rb_define_method(rb_cComplex, "conjugate", rb_complex_conjugate, 0);
     rb_define_method(rb_cComplex, "conj", rb_complex_conjugate, 0);
 
-    rb_define_method(rb_cComplex, "real?", nucomp_false, 0);
+    rb_define_method(rb_cComplex, "real?", nucomp_real_p_m, 0);
 
     rb_define_method(rb_cComplex, "numerator", nucomp_numerator, 0);
     rb_define_method(rb_cComplex, "denominator", nucomp_denominator, 0);
@@ -2430,8 +2388,6 @@ Init_Complex(void)
     rb_define_private_method(compat, "marshal_load", nucomp_marshal_load, 1);
     rb_marshal_define_compat(rb_cComplex, compat, nucomp_dumper, nucomp_loader);
 
-    /* --- */
-
     rb_define_method(rb_cComplex, "to_i", nucomp_to_i, 0);
     rb_define_method(rb_cComplex, "to_f", nucomp_to_f, 0);
     rb_define_method(rb_cComplex, "to_r", nucomp_to_r, 0);
@@ -2444,8 +2400,6 @@ Init_Complex(void)
 
     rb_define_private_method(CLASS_OF(rb_cComplex), "convert", nucomp_s_convert, -1);
 
-    /* --- */
-
     rb_define_method(rb_cNumeric, "real", numeric_real, 0);
     rb_define_method(rb_cNumeric, "imaginary", numeric_imag, 0);
     rb_define_method(rb_cNumeric, "imag", numeric_imag, 0);
