@@ -21,17 +21,15 @@ static VALUE ruby_dln_librefs;
 
 #define IS_RBEXT(e) (strcmp((e), ".rb") == 0)
 #define IS_SOEXT(e) (strcmp((e), ".so") == 0 || strcmp((e), ".o") == 0)
-#ifdef DLEXT2
-#define IS_DLEXT(e) (strcmp((e), DLEXT) == 0 || strcmp((e), DLEXT2) == 0)
-#else
 #define IS_DLEXT(e) (strcmp((e), DLEXT) == 0)
-#endif
 
 static const char *const loadable_ext[] = {
     ".rb", DLEXT,
-#ifdef DLEXT2
-    DLEXT2,
-#endif
+    0
+};
+
+static const char *const ruby_ext[] = {
+    ".rb",
     0
 };
 
@@ -48,9 +46,8 @@ enum expand_type {
    string objects in $LOAD_PATH are frozen.
  */
 static void
-rb_construct_expanded_load_path(enum expand_type type, int *has_relative, int *has_non_cache)
+rb_construct_expanded_load_path(rb_vm_t *vm, enum expand_type type, int *has_relative, int *has_non_cache)
 {
-    rb_vm_t *vm = GET_VM();
     VALUE load_path = vm->load_path;
     VALUE expanded_load_path = vm->expanded_load_path;
     VALUE ary;
@@ -95,16 +92,15 @@ rb_construct_expanded_load_path(enum expand_type type, int *has_relative, int *h
     rb_ary_replace(vm->load_path_snapshot, vm->load_path);
 }
 
-VALUE
-rb_get_expanded_load_path(void)
+static VALUE
+get_expanded_load_path(rb_vm_t *vm)
 {
-    rb_vm_t *vm = GET_VM();
     const VALUE non_cache = Qtrue;
 
     if (!rb_ary_shared_with_p(vm->load_path_snapshot, vm->load_path)) {
 	/* The load path was modified. Rebuild the expanded load path. */
 	int has_relative = 0, has_non_cache = 0;
-	rb_construct_expanded_load_path(EXPAND_ALL, &has_relative, &has_non_cache);
+	rb_construct_expanded_load_path(vm, EXPAND_ALL, &has_relative, &has_non_cache);
 	if (has_relative) {
 	    vm->load_path_check_cache = rb_dir_getwd_ospath();
 	}
@@ -119,7 +115,7 @@ rb_get_expanded_load_path(void)
     else if (vm->load_path_check_cache == non_cache) {
 	int has_relative = 1, has_non_cache = 1;
 	/* Expand only non-cacheable objects. */
-	rb_construct_expanded_load_path(EXPAND_NON_CACHE,
+	rb_construct_expanded_load_path(vm, EXPAND_NON_CACHE,
 					&has_relative, &has_non_cache);
     }
     else if (vm->load_path_check_cache) {
@@ -129,18 +125,24 @@ rb_get_expanded_load_path(void)
 	    /* Current working directory or filesystem encoding was changed.
 	       Expand relative load path and non-cacheable objects again. */
 	    vm->load_path_check_cache = cwd;
-	    rb_construct_expanded_load_path(EXPAND_RELATIVE,
+	    rb_construct_expanded_load_path(vm, EXPAND_RELATIVE,
 					    &has_relative, &has_non_cache);
 	}
 	else {
 	    /* Expand only tilde (User HOME) and non-cacheable objects. */
-	    rb_construct_expanded_load_path(EXPAND_HOME,
+	    rb_construct_expanded_load_path(vm, EXPAND_HOME,
 					    &has_relative, &has_non_cache);
 	}
     }
     return vm->expanded_load_path;
 }
 
+VALUE
+rb_get_expanded_load_path(void)
+{
+    return get_expanded_load_path(GET_VM());
+}
+
 static VALUE
 load_path_getter(ID id, VALUE * p)
 {
@@ -149,34 +151,39 @@ load_path_getter(ID id, VALUE * p)
 }
 
 static VALUE
-get_loaded_features(void)
+get_loaded_features(rb_vm_t *vm)
 {
-    return GET_VM()->loaded_features;
+    return vm->loaded_features;
+}
+
+static VALUE
+get_loaded_features_realpaths(rb_vm_t *vm)
+{
+    return vm->loaded_features_realpaths;
 }
 
 static VALUE
 get_LOADED_FEATURES(ID _x, VALUE *_y)
 {
-    return get_loaded_features();
+    return get_loaded_features(GET_VM());
 }
 
 static void
-reset_loaded_features_snapshot(void)
+reset_loaded_features_snapshot(rb_vm_t *vm)
 {
-    rb_vm_t *vm = GET_VM();
     rb_ary_replace(vm->loaded_features_snapshot, vm->loaded_features);
 }
 
 static struct st_table *
-get_loaded_features_index_raw(void)
+get_loaded_features_index_raw(rb_vm_t *vm)
 {
-    return GET_VM()->loaded_features_index;
+    return vm->loaded_features_index;
 }
 
 static st_table *
-get_loading_table(void)
+get_loading_table(rb_vm_t *vm)
 {
-    return GET_VM()->loading_table;
+    return vm->loading_table;
 }
 
 static st_data_t
@@ -185,34 +192,69 @@ feature_key(const char *str, size_t len)
     return st_hash(str, len, 0xfea7009e);
 }
 
+static bool
+is_rbext_path(VALUE feature_path)
+{
+    long len = RSTRING_LEN(feature_path);
+    long rbext_len = rb_strlen_lit(".rb");
+    if (len <= rbext_len) return false;
+    return IS_RBEXT(RSTRING_PTR(feature_path) + len - rbext_len);
+}
+
 static void
-features_index_add_single(const char* str, size_t len, VALUE offset)
+features_index_add_single(rb_vm_t *vm, const char* str, size_t len, VALUE offset, bool rb)
 {
     struct st_table *features_index;
     VALUE this_feature_index = Qnil;
     st_data_t short_feature_key;
+    st_data_t data;
 
     Check_Type(offset, T_FIXNUM);
     short_feature_key = feature_key(str, len);
 
-    features_index = get_loaded_features_index_raw();
-    st_lookup(features_index, short_feature_key, (st_data_t *)&this_feature_index);
-
-    if (NIL_P(this_feature_index)) {
+    features_index = get_loaded_features_index_raw(vm);
+    if (!st_lookup(features_index, short_feature_key, &data) ||
+        NIL_P(this_feature_index = (VALUE)data)) {
 	st_insert(features_index, short_feature_key, (st_data_t)offset);
     }
-    else if (RB_TYPE_P(this_feature_index, T_FIXNUM)) {
+    else if (FIXNUM_P(this_feature_index)) {
+	VALUE loaded_features = get_loaded_features(vm);
+	VALUE this_feature_path = RARRAY_AREF(loaded_features, FIX2LONG(this_feature_index));
 	VALUE feature_indexes[2];
-	feature_indexes[0] = this_feature_index;
-	feature_indexes[1] = offset;
+	int top = (rb && !is_rbext_path(this_feature_path)) ? 1 : 0;
+	feature_indexes[top^0] = this_feature_index;
+	feature_indexes[top^1] = offset;
 	this_feature_index = (VALUE)xcalloc(1, sizeof(struct RArray));
 	RBASIC(this_feature_index)->flags = T_ARRAY; /* fake VALUE, do not mark/sweep */
 	rb_ary_cat(this_feature_index, feature_indexes, numberof(feature_indexes));
 	st_insert(features_index, short_feature_key, (st_data_t)this_feature_index);
     }
     else {
+        long pos = -1;
+
 	Check_Type(this_feature_index, T_ARRAY);
+        if (rb) {
+            VALUE loaded_features = get_loaded_features(vm);
+            for (long i = 0; i < RARRAY_LEN(this_feature_index); ++i) {
+                VALUE idx = RARRAY_AREF(this_feature_index, i);
+                VALUE this_feature_path = RARRAY_AREF(loaded_features, FIX2LONG(idx));
+                Check_Type(this_feature_path, T_STRING);
+                if (!is_rbext_path(this_feature_path)) {
+                    /* as this_feature_index is a fake VALUE, `push` (which
+                     * doesn't wb_unprotect like as rb_ary_splice) first,
+                     * then rotate partially. */
+                    pos = i;
+                    break;
+                }
+            }
+        }
 	rb_ary_push(this_feature_index, offset);
+        if (pos >= 0) {
+            VALUE *ptr = (VALUE *)RARRAY_CONST_PTR_TRANSIENT(this_feature_index);
+            long len = RARRAY_LEN(this_feature_index);
+            MEMMOVE(ptr + pos, ptr + pos + 1, VALUE, len - pos - 1);
+            ptr[pos] = offset;
+        }
     }
 }
 
@@ -225,9 +267,10 @@ features_index_add_single(const char* str, size_t len, VALUE offset)
    relies on for its fast lookup.
 */
 static void
-features_index_add(VALUE feature, VALUE offset)
+features_index_add(rb_vm_t *vm, VALUE feature, VALUE offset)
 {
     const char *feature_str, *feature_end, *ext, *p;
+    bool rb = false;
 
     feature_str = StringValuePtr(feature);
     feature_end = feature_str + RSTRING_LEN(feature);
@@ -237,6 +280,8 @@ features_index_add(VALUE feature, VALUE offset)
 	    break;
     if (*ext != '.')
 	ext = NULL;
+    else
+        rb = IS_RBEXT(ext);
     /* Now `ext` points to the only string matching %r{^\.[^./]*$} that is
        at the end of `feature`, or is NULL if there is no such string. */
 
@@ -248,14 +293,14 @@ features_index_add(VALUE feature, VALUE offset)
 	if (p < feature_str)
 	    break;
 	/* Now *p == '/'.  We reach this point for every '/' in `feature`. */
-	features_index_add_single(p + 1, feature_end - p - 1, offset);
+	features_index_add_single(vm, p + 1, feature_end - p - 1, offset, false);
 	if (ext) {
-	    features_index_add_single(p + 1, ext - p - 1, offset);
+	    features_index_add_single(vm, p + 1, ext - p - 1, offset, rb);
 	}
     }
-    features_index_add_single(feature_str, feature_end - feature_str, offset);
+    features_index_add_single(vm, feature_str, feature_end - feature_str, offset, false);
     if (ext) {
-	features_index_add_single(feature_str, ext - feature_str, offset);
+	features_index_add_single(vm, feature_str, ext - feature_str, offset, rb);
     }
 }
 
@@ -271,16 +316,18 @@ loaded_features_index_clear_i(st_data_t key, st_data_t val, st_data_t arg)
 }
 
 static st_table *
-get_loaded_features_index(void)
+get_loaded_features_index(rb_vm_t *vm)
 {
     VALUE features;
     int i;
-    rb_vm_t *vm = GET_VM();
 
     if (!rb_ary_shared_with_p(vm->loaded_features_snapshot, vm->loaded_features)) {
 	/* The sharing was broken; something (other than us in rb_provide_feature())
 	   modified loaded_features.  Rebuild the index. */
 	st_foreach(vm->loaded_features_index, loaded_features_index_clear_i, 0);
+
+        VALUE realpaths = vm->loaded_features_realpaths;
+        rb_hash_clear(realpaths);
 	features = vm->loaded_features;
 	for (i = 0; i < RARRAY_LEN(features); i++) {
 	    VALUE entry, as_str;
@@ -289,9 +336,18 @@ get_loaded_features_index(void)
 	    as_str = rb_fstring(rb_str_freeze(as_str));
 	    if (as_str != entry)
 		rb_ary_store(features, i, as_str);
-	    features_index_add(as_str, INT2FIX(i));
+	    features_index_add(vm, as_str, INT2FIX(i));
 	}
-	reset_loaded_features_snapshot();
+	reset_loaded_features_snapshot(vm);
+
+        features = rb_ary_dup(vm->loaded_features_snapshot);
+        long j = RARRAY_LEN(features);
+        for (i = 0; i < j; i++) {
+            VALUE as_str = rb_ary_entry(features, i);
+            VALUE realpath = rb_check_realpath(Qnil, as_str, NULL);
+            if (NIL_P(realpath)) realpath = as_str;
+            rb_hash_aset(realpaths, rb_fstring(realpath), Qtrue);
+        }
     }
     return vm->loaded_features_index;
 }
@@ -372,7 +428,7 @@ loaded_feature_path_i(st_data_t v, st_data_t b, st_data_t f)
 }
 
 static int
-rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const char **fn)
+rb_feature_p(rb_vm_t *vm, const char *feature, const char *ext, int rb, int expanded, const char **fn)
 {
     VALUE features, this_feature_index = Qnil, v, p, load_path = 0;
     const char *f, *e;
@@ -393,11 +449,10 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
 	elen = 0;
 	type = 0;
     }
-    features = get_loaded_features();
-    features_index = get_loaded_features_index();
+    features = get_loaded_features(vm);
+    features_index = get_loaded_features_index(vm);
 
     key = feature_key(feature, strlen(feature));
-    st_lookup(features_index, key, (st_data_t *)&this_feature_index);
     /* We search `features` for an entry such that either
          "#{features[i]}" == "#{load_path[j]}/#{feature}#{e}"
        for some j, or
@@ -424,7 +479,7 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
        or ends in '/'.  This includes both match forms above, as well
        as any distractors, so we may ignore all other entries in `features`.
      */
-    if (!NIL_P(this_feature_index)) {
+    if (st_lookup(features_index, key, &data) && !NIL_P(this_feature_index = (VALUE)data)) {
 	for (i = 0; ; i++) {
 	    VALUE entry;
 	    long index;
@@ -443,7 +498,7 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
 	    if ((n = RSTRING_LEN(v)) < len) continue;
 	    if (strncmp(f, feature, len) != 0) {
 		if (expanded) continue;
-		if (!load_path) load_path = rb_get_expanded_load_path();
+		if (!load_path) load_path = get_expanded_load_path(vm);
 		if (!(p = loaded_feature_path(f, n, feature, len, type, load_path)))
 		    continue;
 		expanded = 1;
@@ -463,14 +518,14 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
 	}
     }
 
-    loading_tbl = get_loading_table();
+    loading_tbl = get_loading_table(vm);
     f = 0;
     if (!expanded) {
 	struct loaded_feature_searching fs;
 	fs.name = feature;
 	fs.len = len;
 	fs.type = type;
-	fs.load_path = load_path ? load_path : rb_get_expanded_load_path();
+	fs.load_path = load_path ? load_path : get_expanded_load_path(vm);
 	fs.result = 0;
 	st_foreach(loading_tbl, loaded_feature_path_i, (st_data_t)&fs);
 	if ((f = fs.result) != 0) {
@@ -524,8 +579,8 @@ rb_provided(const char *feature)
     return rb_feature_provided(feature, 0);
 }
 
-int
-rb_feature_provided(const char *feature, const char **loading)
+static int
+feature_provided(rb_vm_t *vm, const char *feature, const char **loading)
 {
     const char *ext = strrchr(feature, '.');
     VALUE fullpath = 0;
@@ -537,41 +592,48 @@ rb_feature_provided(const char *feature, const char **loading)
     }
     if (ext && !strchr(ext, '/')) {
 	if (IS_RBEXT(ext)) {
-	    if (rb_feature_p(feature, ext, TRUE, FALSE, loading)) return TRUE;
+	    if (rb_feature_p(vm, feature, ext, TRUE, FALSE, loading)) return TRUE;
 	    return FALSE;
 	}
 	else if (IS_SOEXT(ext) || IS_DLEXT(ext)) {
-	    if (rb_feature_p(feature, ext, FALSE, FALSE, loading)) return TRUE;
+	    if (rb_feature_p(vm, feature, ext, FALSE, FALSE, loading)) return TRUE;
 	    return FALSE;
 	}
     }
-    if (rb_feature_p(feature, 0, TRUE, FALSE, loading))
+    if (rb_feature_p(vm, feature, 0, TRUE, FALSE, loading))
 	return TRUE;
     RB_GC_GUARD(fullpath);
     return FALSE;
 }
 
+int
+rb_feature_provided(const char *feature, const char **loading)
+{
+    return feature_provided(GET_VM(), feature, loading);
+}
+
 static void
-rb_provide_feature(VALUE feature)
+rb_provide_feature(rb_vm_t *vm, VALUE feature)
 {
     VALUE features;
 
-    features = get_loaded_features();
+    features = get_loaded_features(vm);
     if (OBJ_FROZEN(features)) {
 	rb_raise(rb_eRuntimeError,
 		 "$LOADED_FEATURES is frozen; cannot append feature");
     }
     rb_str_freeze(feature);
 
+    get_loaded_features_index(vm);
     rb_ary_push(features, rb_fstring(feature));
-    features_index_add(feature, INT2FIX(RARRAY_LEN(features)-1));
-    reset_loaded_features_snapshot();
+    features_index_add(vm, feature, INT2FIX(RARRAY_LEN(features)-1));
+    reset_loaded_features_snapshot(vm);
 }
 
 void
 rb_provide(const char *feature)
 {
-    rb_provide_feature(rb_fstring_cstr(feature));
+    rb_provide_feature(GET_VM(), rb_fstring_cstr(feature));
 }
 
 NORETURN(static void load_failed(VALUE));
@@ -731,10 +793,10 @@ rb_f_load(int argc, VALUE *argv, VALUE _)
 }
 
 static char *
-load_lock(const char *ftptr)
+load_lock(rb_vm_t *vm, const char *ftptr, bool warn)
 {
     st_data_t data;
-    st_table *loading_tbl = get_loading_table();
+    st_table *loading_tbl = get_loading_table(vm);
 
     if (!st_lookup(loading_tbl, (st_data_t)ftptr, &data)) {
 	/* partial state */
@@ -751,7 +813,7 @@ load_lock(const char *ftptr)
 	(*init)();
 	return (char *)"";
     }
-    if (RTEST(ruby_verbose)) {
+    if (warn) {
 	VALUE warning = rb_warning_string("loading in progress, circular require considered harmful - %s", ftptr);
 	rb_backtrace_each(rb_str_append, warning);
 	rb_warning("%"PRIsVALUE, warning);
@@ -783,11 +845,11 @@ release_thread_shield(st_data_t *key, st_data_t *value, st_data_t done, int exis
 }
 
 static void
-load_unlock(const char *ftptr, int done)
+load_unlock(rb_vm_t *vm, const char *ftptr, int done)
 {
     if (ftptr) {
 	st_data_t key = (st_data_t)ftptr;
-	st_table *loading_tbl = get_loading_table();
+	st_table *loading_tbl = get_loading_table(vm);
 
 	st_update(loading_tbl, key, release_thread_shield, done);
     }
@@ -857,10 +919,10 @@ rb_f_require_relative(VALUE obj, VALUE fname)
     return rb_require_string(rb_file_absolute_path(fname, base));
 }
 
-typedef int (*feature_func)(const char *feature, const char *ext, int rb, int expanded, const char **fn);
+typedef int (*feature_func)(rb_vm_t *vm, const char *feature, const char *ext, int rb, int expanded, const char **fn);
 
 static int
-search_required(VALUE fname, volatile VALUE *path, feature_func rb_feature_p)
+search_required(rb_vm_t *vm, VALUE fname, volatile VALUE *path, feature_func rb_feature_p)
 {
     VALUE tmp;
     char *ext, *ftptr;
@@ -871,68 +933,58 @@ search_required(VALUE fname, volatile VALUE *path, feature_func rb_feature_p)
     ext = strrchr(ftptr = RSTRING_PTR(fname), '.');
     if (ext && !strchr(ext, '/')) {
 	if (IS_RBEXT(ext)) {
-	    if (rb_feature_p(ftptr, ext, TRUE, FALSE, &loading)) {
+	    if (rb_feature_p(vm, ftptr, ext, TRUE, FALSE, &loading)) {
 		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 'r';
 	    }
             if ((tmp = rb_find_file(fname)) != 0) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
-		if (!rb_feature_p(ftptr, ext, TRUE, TRUE, &loading) || loading)
+		if (!rb_feature_p(vm, ftptr, ext, TRUE, TRUE, &loading) || loading)
 		    *path = tmp;
 		return 'r';
 	    }
 	    return 0;
 	}
 	else if (IS_SOEXT(ext)) {
-	    if (rb_feature_p(ftptr, ext, FALSE, FALSE, &loading)) {
+	    if (rb_feature_p(vm, ftptr, ext, FALSE, FALSE, &loading)) {
 		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 's';
 	    }
 	    tmp = rb_str_subseq(fname, 0, ext - RSTRING_PTR(fname));
-#ifdef DLEXT2
-	    OBJ_FREEZE(tmp);
-            if (rb_find_file_ext(&tmp, loadable_ext + 1)) {
-		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
-		if (!rb_feature_p(ftptr, ext, FALSE, TRUE, &loading) || loading)
-		    *path = tmp;
-		return 's';
-	    }
-#else
 	    rb_str_cat2(tmp, DLEXT);
 	    OBJ_FREEZE(tmp);
             if ((tmp = rb_find_file(tmp)) != 0) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
-		if (!rb_feature_p(ftptr, ext, FALSE, TRUE, &loading) || loading)
+		if (!rb_feature_p(vm, ftptr, ext, FALSE, TRUE, &loading) || loading)
 		    *path = tmp;
 		return 's';
 	    }
-#endif
 	}
 	else if (IS_DLEXT(ext)) {
-	    if (rb_feature_p(ftptr, ext, FALSE, FALSE, &loading)) {
+	    if (rb_feature_p(vm, ftptr, ext, FALSE, FALSE, &loading)) {
 		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 's';
 	    }
             if ((tmp = rb_find_file(fname)) != 0) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
-		if (!rb_feature_p(ftptr, ext, FALSE, TRUE, &loading) || loading)
+		if (!rb_feature_p(vm, ftptr, ext, FALSE, TRUE, &loading) || loading)
 		    *path = tmp;
 		return 's';
 	    }
 	}
     }
-    else if ((ft = rb_feature_p(ftptr, 0, FALSE, FALSE, &loading)) == 'r') {
+    else if ((ft = rb_feature_p(vm, ftptr, 0, FALSE, FALSE, &loading)) == 'r') {
 	if (loading) *path = rb_filesystem_str_new_cstr(loading);
 	return 'r';
     }
     tmp = fname;
-    type = rb_find_file_ext(&tmp, loadable_ext);
+    type = rb_find_file_ext(&tmp, ft == 's' ? ruby_ext : loadable_ext);
     switch (type) {
       case 0:
 	if (ft)
 	    goto statically_linked;
 	ftptr = RSTRING_PTR(tmp);
-	return rb_feature_p(ftptr, 0, FALSE, TRUE, 0);
+	return rb_feature_p(vm, ftptr, 0, FALSE, TRUE, 0);
 
       default:
 	if (ft) {
@@ -941,7 +993,7 @@ search_required(VALUE fname, volatile VALUE *path, feature_func rb_feature_p)
         /* fall through */
       case 1:
 	ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
-	if (rb_feature_p(ftptr, ext, !--type, TRUE, &loading) && !loading)
+	if (rb_feature_p(vm, ftptr, ext, !--type, TRUE, &loading) && !loading)
 	    break;
 	*path = tmp;
     }
@@ -966,7 +1018,7 @@ load_ext(VALUE path)
 }
 
 static int
-no_feature_p(const char *feature, const char *ext, int rb, int expanded, const char **fn)
+no_feature_p(rb_vm_t *vm, const char *feature, const char *ext, int rb, int expanded, const char **fn)
 {
     return 0;
 }
@@ -981,7 +1033,7 @@ rb_resolve_feature_path(VALUE klass, VALUE fname)
 
     fname = rb_get_path(fname);
     path = rb_str_encode_ospath(fname);
-    found = search_required(path, &path, no_feature_p);
+    found = search_required(GET_VM(), path, &path, no_feature_p);
 
     switch (found) {
       case 'r':
@@ -991,12 +1043,31 @@ rb_resolve_feature_path(VALUE klass, VALUE fname)
         sym = ID2SYM(rb_intern("so"));
         break;
       default:
-        load_failed(fname);
+        return Qnil;
     }
 
     return rb_ary_new_from_args(2, sym, path);
 }
 
+static void
+ext_config_push(rb_thread_t *th, struct rb_ext_config *prev)
+{
+    *prev = th->ext_config;
+    th->ext_config = (struct rb_ext_config){0};
+}
+
+static void
+ext_config_pop(rb_thread_t *th, struct rb_ext_config *prev)
+{
+    th->ext_config = *prev;
+}
+
+void
+rb_ext_ractor_safe(bool flag)
+{
+    GET_THREAD()->ext_config.ractor_safe = flag;
+}
+
 /*
  * returns
  *  0: if already loaded (false)
@@ -1005,20 +1076,28 @@ rb_resolve_feature_path(VALUE klass, VALUE fname)
  * >1: exception
  */
 static int
-require_internal(rb_execution_context_t *ec, VALUE fname, int exception)
+require_internal(rb_execution_context_t *ec, VALUE fname, int exception, bool warn)
 {
     volatile int result = -1;
     rb_thread_t *th = rb_ec_thread_ptr(ec);
-    volatile VALUE wrapper = th->top_wrapper;
-    volatile VALUE self = th->top_self;
-    volatile VALUE errinfo = ec->errinfo;
+    volatile const struct {
+        VALUE wrapper, self, errinfo;
+    } saved = {
+        th->top_wrapper, th->top_self, ec->errinfo,
+    };
     enum ruby_tag_type state;
     char *volatile ftptr = 0;
     VALUE path;
+    volatile VALUE saved_path;
+    volatile VALUE realpath = 0;
+    VALUE realpaths = get_loaded_features_realpaths(th->vm);
+    volatile bool reset_ext_config = false;
+    struct rb_ext_config prev_ext_config;
 
     fname = rb_get_path(fname);
     path = rb_str_encode_ospath(fname);
     RUBY_DTRACE_HOOK(REQUIRE_ENTRY, RSTRING_PTR(fname));
+    saved_path = path;
 
     EC_PUSH_TAG(ec);
     ec->errinfo = Qnil; /* ensure */
@@ -1028,16 +1107,21 @@ require_internal(rb_execution_context_t *ec, VALUE fname, int exception)
 	int found;
 
 	RUBY_DTRACE_HOOK(FIND_REQUIRE_ENTRY, RSTRING_PTR(fname));
-        found = search_required(path, &path, rb_feature_p);
+        found = search_required(th->vm, path, &saved_path, rb_feature_p);
 	RUBY_DTRACE_HOOK(FIND_REQUIRE_RETURN, RSTRING_PTR(fname));
+        path = saved_path;
 
 	if (found) {
-            if (!path || !(ftptr = load_lock(RSTRING_PTR(path)))) {
+            if (!path || !(ftptr = load_lock(th->vm, RSTRING_PTR(path), warn))) {
 		result = 0;
 	    }
 	    else if (!*ftptr) {
 		result = TAG_RETURN;
 	    }
+            else if (RTEST(rb_hash_aref(realpaths,
+                                        realpath = rb_realpath_internal(Qnil, path, 1)))) {
+                result = 0;
+            }
 	    else {
 		switch (found) {
 		  case 'r':
@@ -1045,6 +1129,8 @@ require_internal(rb_execution_context_t *ec, VALUE fname, int exception)
 		    break;
 
 		  case 's':
+                    reset_ext_config = true;
+                    ext_config_push(th, &prev_ext_config);
 		    handle = (long)rb_vm_call_cfunc(rb_vm_top_self(), load_ext,
 						    path, VM_BLOCK_HANDLER_NONE, path);
 		    rb_ary_push(ruby_dln_librefs, LONG2NUM(handle));
@@ -1055,10 +1141,14 @@ require_internal(rb_execution_context_t *ec, VALUE fname, int exception)
 	}
     }
     EC_POP_TAG();
-    th = rb_ec_thread_ptr(ec);
-    th->top_self = self;
-    th->top_wrapper = wrapper;
-    if (ftptr) load_unlock(RSTRING_PTR(path), !state);
+
+    rb_thread_t *th2 = rb_ec_thread_ptr(ec);
+    th2->top_self = saved.self;
+    th2->top_wrapper = saved.wrapper;
+    if (reset_ext_config) ext_config_pop(th2, &prev_ext_config);
+
+    path = saved_path;
+    if (ftptr) load_unlock(th2->vm, RSTRING_PTR(path), !state);
 
     if (state) {
         if (state == TAG_FATAL) {
@@ -1083,19 +1173,32 @@ require_internal(rb_execution_context_t *ec, VALUE fname, int exception)
         rb_exc_raise(ec->errinfo);
     }
 
-    if (result == TAG_RETURN) rb_provide_feature(path);
-    ec->errinfo = errinfo;
+    if (result == TAG_RETURN) {
+        rb_provide_feature(th2->vm, path);
+        VALUE real = realpath;
+        if (real) {
+            rb_hash_aset(realpaths, rb_fstring(real), Qtrue);
+        }
+    }
+    ec->errinfo = saved.errinfo;
 
     RUBY_DTRACE_HOOK(REQUIRE_RETURN, RSTRING_PTR(fname));
 
     return result;
 }
 
+int
+rb_require_internal_silent(VALUE fname)
+{
+    rb_execution_context_t *ec = GET_EC();
+    return require_internal(ec, fname, 1, false);
+}
+
 int
 rb_require_internal(VALUE fname)
 {
     rb_execution_context_t *ec = GET_EC();
-    return require_internal(ec, fname, 1);
+    return require_internal(ec, fname, 1, RTEST(ruby_verbose));
 }
 
 int
@@ -1104,7 +1207,7 @@ ruby_require_internal(const char *fname, unsigned int len)
     struct RString fake;
     VALUE str = rb_setup_fake_str(&fake, fname, len, 0);
     rb_execution_context_t *ec = GET_EC();
-    int result = require_internal(ec, str, 0);
+    int result = require_internal(ec, str, 0, RTEST(ruby_verbose));
     rb_set_errinfo(Qnil);
     return result == TAG_RETURN ? 1 : result ? -1 : 0;
 }
@@ -1113,7 +1216,7 @@ VALUE
 rb_require_string(VALUE fname)
 {
     rb_execution_context_t *ec = GET_EC();
-    int result = require_internal(ec, fname, 1);
+    int result = require_internal(ec, fname, 1, RTEST(ruby_verbose));
 
     if (result > TAG_RETURN) {
         EC_JUMP_TAG(ec, result);
@@ -1122,7 +1225,7 @@ rb_require_string(VALUE fname)
 	load_failed(fname);
     }
 
-    return result ? Qtrue : Qfalse;
+    return RBOOL(result);
 }
 
 VALUE
@@ -1149,9 +1252,10 @@ register_init_ext(st_data_t *key, st_data_t *value, st_data_t init, int existing
 RUBY_FUNC_EXPORTED void
 ruby_init_ext(const char *name, void (*init)(void))
 {
-    st_table *loading_tbl = get_loading_table();
+    rb_vm_t *vm = GET_VM();
+    st_table *loading_tbl = get_loading_table(vm);
 
-    if (rb_provided(name))
+    if (feature_provided(vm, name, 0))
 	return;
     st_update(loading_tbl, (st_data_t)name, register_init_ext, (st_data_t)init);
 }
@@ -1234,7 +1338,7 @@ static VALUE
 rb_f_autoload(VALUE obj, VALUE sym, VALUE file)
 {
     VALUE klass = rb_class_real(rb_vm_cbase());
-    if (NIL_P(klass)) {
+    if (!klass) {
 	rb_raise(rb_eTypeError, "Can not set autoload on singleton class");
     }
     return rb_mod_autoload(klass, sym, file);
@@ -1265,15 +1369,13 @@ rb_f_autoload_p(int argc, VALUE *argv, VALUE obj)
 void
 Init_load(void)
 {
-#undef rb_intern
-#define rb_intern(str) rb_intern2((str), strlen(str))
     rb_vm_t *vm = GET_VM();
     static const char var_load_path[] = "$:";
     ID id_load_path = rb_intern2(var_load_path, sizeof(var_load_path)-1);
 
     rb_define_hooked_variable(var_load_path, (VALUE*)vm, load_path_getter, rb_gvar_readonly_setter);
-    rb_alias_variable(rb_intern("$-I"), id_load_path);
-    rb_alias_variable(rb_intern("$LOAD_PATH"), id_load_path);
+    rb_alias_variable(rb_intern_const("$-I"), id_load_path);
+    rb_alias_variable(rb_intern_const("$LOAD_PATH"), id_load_path);
     vm->load_path = rb_ary_new();
     vm->expanded_load_path = rb_ary_tmp_new(0);
     vm->load_path_snapshot = rb_ary_tmp_new(0);
@@ -1285,6 +1387,8 @@ Init_load(void)
     vm->loaded_features = rb_ary_new();
     vm->loaded_features_snapshot = rb_ary_tmp_new(0);
     vm->loaded_features_index = st_init_numtable();
+    vm->loaded_features_realpaths = rb_hash_new();
+    rb_obj_hide(vm->loaded_features_realpaths);
 
     rb_define_global_function("load", rb_f_load, -1);
     rb_define_global_function("require", rb_f_require, 1);
