@@ -20,8 +20,6 @@
 #include <gmp.h>
 #endif
 
-#undef NDEBUG
-#define NDEBUG
 #include "id.h"
 #include "internal.h"
 #include "internal/array.h"
@@ -48,11 +46,11 @@ static ID id_abs, id_integer_p,
 #define id_idiv idDiv
 #define id_to_i idTo_i
 
-#define f_boolcast(x) ((x) ? Qtrue : Qfalse)
 #define f_inspect rb_inspect
 #define f_to_s rb_obj_as_string
 
 static VALUE nurat_to_f(VALUE self);
+static VALUE float_to_r(VALUE self);
 
 inline static VALUE
 f_add(VALUE x, VALUE y)
@@ -130,7 +128,7 @@ f_abs(VALUE x)
 }
 
 
-inline static VALUE
+inline static int
 f_integer_p(VALUE x)
 {
     return RB_INTEGER_TYPE_P(x);
@@ -201,7 +199,7 @@ f_minus_one_p(VALUE x)
     if (RB_INTEGER_TYPE_P(x)) {
 	return x == LONG2FIX(-1);
     }
-    else if (RB_TYPE_P(x, T_BIGNUM)) {
+    else if (RB_BIGNUM_TYPE_P(x)) {
 	return Qfalse;
     }
     else if (RB_TYPE_P(x, T_RATIONAL)) {
@@ -367,7 +365,7 @@ inline static VALUE
 f_gcd(VALUE x, VALUE y)
 {
 #ifdef USE_GMP
-    if (RB_TYPE_P(x, T_BIGNUM) && RB_TYPE_P(y, T_BIGNUM)) {
+    if (RB_BIGNUM_TYPE_P(x) && RB_BIGNUM_TYPE_P(y)) {
         size_t xn = BIGNUM_LEN(x);
         size_t yn = BIGNUM_LEN(y);
         if (GMP_GCD_DIGITS <= xn || GMP_GCD_DIGITS <= yn)
@@ -430,18 +428,6 @@ f_rational_new_bang1(VALUE klass, VALUE x)
     return nurat_s_new_internal(klass, x, ONE);
 }
 
-#ifdef CANONICALIZATION_FOR_MATHN
-static int canonicalization = 0;
-
-RUBY_FUNC_EXPORTED void
-nurat_canonicalization(int f)
-{
-    canonicalization = f;
-}
-#else
-# define canonicalization 0
-#endif
-
 inline static void
 nurat_int_check(VALUE num)
 {
@@ -490,8 +476,6 @@ nurat_s_canonicalize_internal(VALUE klass, VALUE num, VALUE den)
     nurat_canonicalize(&num, &den);
     nurat_reduce(&num, &den);
 
-    if (canonicalization && f_one_p(den))
-	return num;
     return nurat_s_new_internal(klass, num, den);
 }
 
@@ -500,8 +484,6 @@ nurat_s_canonicalize_internal_no_reduce(VALUE klass, VALUE num, VALUE den)
 {
     nurat_canonicalize(&num, &den);
 
-    if (canonicalization && f_one_p(den))
-	return num;
     return nurat_s_new_internal(klass, num, den);
 }
 
@@ -910,8 +892,8 @@ rb_rational_mul(VALUE self, VALUE other)
  *    Rational(9, 8)  / 4                #=> (9/32)
  *    Rational(20, 9) / 9.8              #=> 0.22675736961451246
  */
-static VALUE
-nurat_div(VALUE self, VALUE other)
+VALUE
+rb_rational_div(VALUE self, VALUE other)
 {
     if (RB_INTEGER_TYPE_P(other)) {
 	if (f_zero_p(other))
@@ -963,10 +945,10 @@ nurat_fdiv(VALUE self, VALUE other)
 {
     VALUE div;
     if (f_zero_p(other))
-        return nurat_div(self, rb_float_new(0.0));
+        return rb_rational_div(self, rb_float_new(0.0));
     if (FIXNUM_P(other) && other == LONG2FIX(1))
 	return nurat_to_f(self);
-    div = nurat_div(self, other);
+    div = rb_rational_div(self, other);
     if (RB_TYPE_P(div, T_RATIONAL))
 	return nurat_to_f(div);
     if (RB_FLOAT_TYPE_P(div))
@@ -1052,7 +1034,7 @@ rb_rational_pow(VALUE self, VALUE other)
 	    return f_rational_new2(CLASS_OF(self), num, den);
 	}
     }
-    else if (RB_TYPE_P(other, T_BIGNUM)) {
+    else if (RB_BIGNUM_TYPE_P(other)) {
 	rb_warn("in a**b, b may be too big");
 	return rb_float_pow(nurat_to_f(self), other);
     }
@@ -1060,7 +1042,7 @@ rb_rational_pow(VALUE self, VALUE other)
 	return rb_float_pow(nurat_to_f(self), other);
     }
     else {
-	return rb_num_coerce_bin(self, other, rb_intern("**"));
+	return rb_num_coerce_bin(self, other, idPow);
     }
 }
 #define nurat_expt rb_rational_pow
@@ -1119,7 +1101,7 @@ rb_rational_cmp(VALUE self, VALUE other)
         return rb_dbl_cmp(nurat_to_double(self), RFLOAT_VALUE(other));
 
       default:
-	return rb_num_coerce_cmp(self, other, rb_intern("<=>"));
+	return rb_num_coerce_cmp(self, other, idCmp);
     }
 }
 
@@ -1153,12 +1135,12 @@ nurat_eqeq_p(VALUE self, VALUE other)
 	}
         else {
             const double d = nurat_to_double(self);
-            return f_boolcast(FIXNUM_ZERO_P(rb_dbl_cmp(d, NUM2DBL(other))));
+            return RBOOL(FIXNUM_ZERO_P(rb_dbl_cmp(d, NUM2DBL(other))));
         }
     }
     else if (RB_FLOAT_TYPE_P(other)) {
 	const double d = nurat_to_double(self);
-	return f_boolcast(FIXNUM_ZERO_P(rb_dbl_cmp(d, RFLOAT_VALUE(other))));
+	return RBOOL(FIXNUM_ZERO_P(rb_dbl_cmp(d, RFLOAT_VALUE(other))));
     }
     else if (RB_TYPE_P(other, T_RATIONAL)) {
 	{
@@ -1167,7 +1149,7 @@ nurat_eqeq_p(VALUE self, VALUE other)
 	    if (INT_ZERO_P(adat->num) && INT_ZERO_P(bdat->num))
 		return Qtrue;
 
-	    return f_boolcast(rb_int_equal(adat->num, bdat->num) &&
+	    return RBOOL(rb_int_equal(adat->num, bdat->num) &&
 			      rb_int_equal(adat->den, bdat->den));
 	}
     }
@@ -1190,11 +1172,17 @@ nurat_coerce(VALUE self, VALUE other)
 	return rb_assoc_new(other, self);
     }
     else if (RB_TYPE_P(other, T_COMPLEX)) {
-	if (k_exact_zero_p(RCOMPLEX(other)->imag))
-	    return rb_assoc_new(f_rational_new_bang1
-				(CLASS_OF(self), RCOMPLEX(other)->real), self);
-	else
+	if (!k_exact_zero_p(RCOMPLEX(other)->imag))
 	    return rb_assoc_new(other, rb_Complex(self, INT2FIX(0)));
+        other = RCOMPLEX(other)->real;
+        if (RB_FLOAT_TYPE_P(other)) {
+            other = float_to_r(other);
+            RBASIC_SET_CLASS(other, CLASS_OF(self));
+        }
+        else {
+            other = f_rational_new_bang1(CLASS_OF(self), other);
+        }
+        return rb_assoc_new(other, self);
     }
 
     rb_raise(rb_eTypeError, "%s can't be coerced into %s",
@@ -1212,7 +1200,7 @@ static VALUE
 nurat_positive_p(VALUE self)
 {
     get_dat1(self);
-    return f_boolcast(INT_POSITIVE_P(dat->num));
+    return RBOOL(INT_POSITIVE_P(dat->num));
 }
 
 /*
@@ -1225,7 +1213,7 @@ static VALUE
 nurat_negative_p(VALUE self)
 {
     get_dat1(self);
-    return f_boolcast(INT_NEGATIVE_P(dat->num));
+    return RBOOL(INT_NEGATIVE_P(dat->num));
 }
 
 /*
@@ -1393,7 +1381,7 @@ f_round_common(int argc, VALUE *argv, VALUE self, VALUE (*func)(VALUE))
 
     s = (*func)(s);
 
-    s = nurat_div(f_rational_new_bang1(CLASS_OF(self), s), b);
+    s = rb_rational_div(f_rational_new_bang1(CLASS_OF(self), s), b);
 
     if (RB_TYPE_P(s, T_RATIONAL) && FIX2INT(rb_int_cmp(n, ONE)) < 0)
 	s = nurat_truncate(s);
@@ -1401,6 +1389,18 @@ f_round_common(int argc, VALUE *argv, VALUE self, VALUE (*func)(VALUE))
     return s;
 }
 
+VALUE
+rb_rational_floor(VALUE self, int ndigits)
+{
+    if (ndigits == 0) {
+        return nurat_floor(self);
+    }
+    else {
+        VALUE n = INT2NUM(ndigits);
+        return f_round_common(1, &n, self, nurat_floor);
+    }
+}
+
 /*
  * call-seq:
  *    rat.floor([ndigits])  ->  integer or rational
@@ -1539,6 +1539,12 @@ nurat_round_n(int argc, VALUE *argv, VALUE self)
     return f_round_common(argc, argv, self, round_func);
 }
 
+VALUE
+rb_flo_round_by_rational(int argc, VALUE *argv, VALUE num)
+{
+    return nurat_to_f(nurat_round_n(argc, argv, float_to_r(num)));
+}
+
 static double
 nurat_to_double(VALUE self)
 {
@@ -1741,8 +1747,8 @@ nurat_rationalize(int argc, VALUE *argv, VALUE self)
 }
 
 /* :nodoc: */
-static VALUE
-nurat_hash(VALUE self)
+st_index_t
+rb_rational_hash(VALUE self)
 {
     st_index_t v, h[2];
     VALUE n;
@@ -1753,9 +1759,16 @@ nurat_hash(VALUE self)
     n = rb_hash(dat->den);
     h[1] = NUM2LONG(n);
     v = rb_memhash(h, sizeof(h));
-    return ST2FIX(v);
+    return v;
 }
 
+static VALUE
+nurat_hash(VALUE self)
+{
+    return ST2FIX(rb_rational_hash(self));
+}
+
+
 static VALUE
 f_format(VALUE self, VALUE (*func)(VALUE))
 {
@@ -1868,13 +1881,11 @@ nurat_marshal_load(VALUE self, VALUE a)
     return self;
 }
 
-/* --- */
-
 VALUE
 rb_rational_reciprocal(VALUE x)
 {
     get_dat1(x);
-    return f_rational_new_no_reduce2(CLASS_OF(x), dat->den, dat->num);
+    return nurat_convert(CLASS_OF(x), dat->den, dat->num, FALSE);
 }
 
 /*
@@ -2028,13 +2039,8 @@ rb_numeric_quo(VALUE x, VALUE y)
         return rb_funcallv(x, idFdiv, 1, &y);
     }
 
-    if (canonicalization) {
-        x = rb_rational_raw1(x);
-    }
-    else {
-        x = rb_convert_type(x, T_RATIONAL, "Rational", "to_r");
-    }
-    return nurat_div(x, y);
+    x = rb_convert_type(x, T_RATIONAL, "Rational", "to_r");
+    return rb_rational_div(x, y);
 }
 
 VALUE
@@ -2071,7 +2077,6 @@ integer_denominator(VALUE self)
     return INT2FIX(1);
 }
 
-static VALUE float_to_r(VALUE self);
 /*
  * call-seq:
  *    flo.numerator  ->  integer
@@ -2089,12 +2094,9 @@ rb_float_numerator(VALUE self)
 {
     double d = RFLOAT_VALUE(self);
     VALUE r;
-    if (isinf(d) || isnan(d))
+    if (!isfinite(d))
 	return self;
     r = float_to_r(self);
-    if (canonicalization && k_integer_p(r)) {
-	return r;
-    }
     return nurat_numerator(r);
 }
 
@@ -2112,12 +2114,9 @@ rb_float_denominator(VALUE self)
 {
     double d = RFLOAT_VALUE(self);
     VALUE r;
-    if (isinf(d) || isnan(d))
+    if (!isfinite(d))
 	return INT2FIX(1);
     r = float_to_r(self);
-    if (canonicalization && k_integer_p(r)) {
-	return ONE;
-    }
     return nurat_denominator(r);
 }
 
@@ -2425,7 +2424,7 @@ parse_rat(const char *s, const char *const e, int strict, int raise)
 
     if (!read_num(&s, e, &num, &nexp)) {
 	if (strict) return Qnil;
-	return canonicalization ? ZERO : nurat_s_alloc(rb_cRational);
+	return nurat_s_alloc(rb_cRational);
     }
     den = ONE;
     if (s < e && *s == '/') {
@@ -2481,9 +2480,7 @@ parse_rat(const char *s, const char *const e, int strict, int raise)
 	num = negate_num(num);
     }
 
-    if (!canonicalization || den != ONE)
-	num = rb_rational_raw(num, den);
-    return num;
+    return rb_rational_raw(num, den);
 }
 
 static VALUE
@@ -2573,6 +2570,8 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
     VALUE a1 = numv, a2 = denv;
     int state;
 
+    assert(a1 != Qundef);
+
     if (NIL_P(a1) || NIL_P(a2)) {
         if (!raise) return Qnil;
         rb_raise(rb_eTypeError, "can't convert nil into Rational");
@@ -2588,21 +2587,47 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
             a2 = RCOMPLEX(a2)->real;
     }
 
-    if (RB_FLOAT_TYPE_P(a1)) {
+    if (RB_INTEGER_TYPE_P(a1)) {
+        // nothing to do
+    }
+    else if (RB_FLOAT_TYPE_P(a1)) {
         a1 = float_to_r(a1);
     }
+    else if (RB_TYPE_P(a1, T_RATIONAL)) {
+        // nothing to do
+    }
     else if (RB_TYPE_P(a1, T_STRING)) {
         a1 = string_to_r_strict(a1, raise);
         if (!raise && NIL_P(a1)) return Qnil;
     }
+    else if (!rb_respond_to(a1, idTo_r)) {
+        VALUE tmp = rb_protect(rb_check_to_int, a1, NULL);
+        rb_set_errinfo(Qnil);
+        if (!NIL_P(tmp)) {
+            a1 = tmp;
+        }
+    }
 
-    if (RB_FLOAT_TYPE_P(a2)) {
+    if (RB_INTEGER_TYPE_P(a2)) {
+        // nothing to do
+    }
+    else if (RB_FLOAT_TYPE_P(a2)) {
         a2 = float_to_r(a2);
     }
+    else if (RB_TYPE_P(a2, T_RATIONAL)) {
+        // nothing to do
+    }
     else if (RB_TYPE_P(a2, T_STRING)) {
         a2 = string_to_r_strict(a2, raise);
         if (!raise && NIL_P(a2)) return Qnil;
     }
+    else if (a2 != Qundef && !rb_respond_to(a2, idTo_r)) {
+        VALUE tmp = rb_protect(rb_check_to_int, a2, NULL);
+        rb_set_errinfo(Qnil);
+        if (!NIL_P(tmp)) {
+            a2 = tmp;
+        }
+    }
 
     if (RB_TYPE_P(a1, T_RATIONAL)) {
         if (a2 == Qundef || (k_exact_one_p(a2)))
@@ -2610,7 +2635,7 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
     }
 
     if (a2 == Qundef) {
-        if (!k_integer_p(a1)) {
+        if (!RB_INTEGER_TYPE_P(a1)) {
             if (!raise) {
                 VALUE result = rb_protect(to_rational, a1, NULL);
                 rb_set_errinfo(Qnil);
@@ -2645,8 +2670,16 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
             }
         }
         if ((k_numeric_p(a1) && k_numeric_p(a2)) &&
-                (!f_integer_p(a1) || !f_integer_p(a2)))
+                (!f_integer_p(a1) || !f_integer_p(a2))) {
+            VALUE tmp = rb_protect(to_rational, a1, &state);
+            if (!state) {
+                a1 = tmp;
+            }
+            else {
+                rb_set_errinfo(Qnil);
+            }
             return f_div(a1, a2);
+        }
     }
 
     a1 = nurat_int_value(a1);
@@ -2723,13 +2756,10 @@ void
 Init_Rational(void)
 {
     VALUE compat;
-#undef rb_intern
-#define rb_intern(str) rb_intern_const(str)
-
-    id_abs = rb_intern("abs");
-    id_integer_p = rb_intern("integer?");
-    id_i_num = rb_intern("@numerator");
-    id_i_den = rb_intern("@denominator");
+    id_abs = rb_intern_const("abs");
+    id_integer_p = rb_intern_const("integer?");
+    id_i_num = rb_intern_const("@numerator");
+    id_i_den = rb_intern_const("@denominator");
 
     rb_cRational = rb_define_class("Rational", rb_cNumeric);
 
@@ -2747,8 +2777,8 @@ Init_Rational(void)
     rb_define_method(rb_cRational, "+", rb_rational_plus, 1);
     rb_define_method(rb_cRational, "-", rb_rational_minus, 1);
     rb_define_method(rb_cRational, "*", rb_rational_mul, 1);
-    rb_define_method(rb_cRational, "/", nurat_div, 1);
-    rb_define_method(rb_cRational, "quo", nurat_div, 1);
+    rb_define_method(rb_cRational, "/", rb_rational_div, 1);
+    rb_define_method(rb_cRational, "quo", rb_rational_div, 1);
     rb_define_method(rb_cRational, "fdiv", nurat_fdiv, 1);
     rb_define_method(rb_cRational, "**", nurat_expt, 1);
 
@@ -2782,8 +2812,6 @@ Init_Rational(void)
     rb_define_private_method(compat, "marshal_load", nurat_marshal_load, 1);
     rb_marshal_define_compat(rb_cRational, compat, nurat_dumper, nurat_loader);
 
-    /* --- */
-
     rb_define_method(rb_cInteger, "gcd", rb_gcd, 1);
     rb_define_method(rb_cInteger, "lcm", rb_lcm, 1);
     rb_define_method(rb_cInteger, "gcdlcm", rb_gcdlcm, 1);
